#include <SPI.h>
#include <Adafruit_TLC59711.h>

#define LED_EN_PIN 4

#define BUZZER_PIN 14
#define BUZZER_PWM_CH 1

#define VIBRATION_PIN 32
#define VIBRATION_PWM_CH 2

#define BUTTON_PIN 25

#define VBUS_ADC_PIN 35
#define VBATT_ADC_PIN 33
#define CHARGING_PIN 16

SPIClass vspi(VSPI);                // Use the VSPI interface
Adafruit_TLC59711 leds(1, &vspi);   // 1 IC

// Wether the user pressed the button to interrupt the session
int INTE = 0;
/**
 * Provide interruption functionality by setting at 1 the global interrupt VAR
 */
void interr() {
    Serial.println("called interrupt");
    INTE = 1;
}

//------------------------- Noise Loop Config Var -------------------------//

// Time interval between itch simulations
const int TIME_INTERVAL = 30 * 1000;

/**
 * Multiple itch simulation in a single noise loop with: 
 * LENGTH as session duration
 * TIME_INTERVAL as pause between itch_simulation
 */
void noise_loop();

//------------------------- Itch simulation Config Var -------------------------//

// Duration of a single itch simulation
const int ITCH_SIM_DURATION = 15 * 1000;
// Itch simulation track (record) to provide increasing / descreasing itch sensation inside a single itch simulation
const int ITCH_SIM_TRACK_LEN = 10;
const int VIBRATIONS[ITCH_SIM_TRACK_LEN] = {160, 160, 185, 160, 185, 210, 185, 185, 210, 210};
const int BUZZERS[ITCH_SIM_TRACK_LEN] = {0, 0, 0, 1, 0, 0, 0, 1, 0, 1};

// flag to enable / disable VIBRATION
const int VIBRATION = 1; 
// flag to enable / disable BUZZER
const int BUZZER = 0;
// flag to enable / disable LEDS
const int LED = 1;
// Backgroud vibation to simulata an always present itch
const int BACKGROUND_VIBRATION = 0;

/**
 * Simulate a single itch sensation inside a session, 
 * itch can be simulated with:
 * VIBRATION, BUZZER and LEDS
 */
void itch_simulation();

//-----------------------------------------------------------------------------//

/**
 * Battery management utils
 */
int adc_to_mv(int r1, int r2, uint32_t value);


void setup() {
    pinMode(LED_EN_PIN, OUTPUT);
    digitalWrite(LED_EN_PIN, LOW);  // Turn on the TLC59711
    delay(10);                      // Wait for it...

  
    leds.setBrightness(40);
    leds.begin();
    leds.write();
  
    ledcAttachPin(BUZZER_PIN, BUZZER_PWM_CH);
    ledcSetup(BUZZER_PWM_CH, 2700, 1);

    ledcAttachPin(VIBRATION_PIN, VIBRATION_PWM_CH);
    ledcSetup(VIBRATION_PWM_CH, 10000, 8);      //10 kHz 8-bit PWM (256 intensity levels)
 
    pinMode(BUTTON_PIN, INPUT);

    pinMode(VBUS_ADC_PIN, INPUT);
    pinMode(CHARGING_PIN, INPUT_PULLUP);

    analogReadResolution(10);
    analogSetAttenuation(ADC_11db);
    
    attachInterrupt(digitalPinToInterrupt(BUTTON_PIN), interr, HIGH);
    
    Serial.begin(115200);
}

void loop() {
    // Small delay before setting at 0 interruption variable, this avoid overwrite by interr function 
    delay(5000);
    INTE = 0;
    
    Serial.println("Hello world");

    noise_loop();

    for (int i = 0; i < 3; i++) {
        if (digitalRead(VBUS_ADC_PIN) == LOW)
            Serial.println("Not charging, insert USB to charge battery");
        else {
            if (digitalRead(CHARGING_PIN) == LOW)
                Serial.println("Battery charge in progress...");
            else
                Serial.println("Charge completed!");
        }

        Serial.printf("Battery voltage: %d mV\n", adc_to_mv(270, 330, analogRead(VBATT_ADC_PIN)));

        delay(1000);
    }

    INTE = 0;
    
    esp_sleep_enable_ext0_wakeup(gpio_num_t(BUTTON_PIN), 0);

    Serial.println("Going to sleep, press the button to restart");
    digitalWrite(LED_EN_PIN, HIGH);      // Turn off the TLC59711


    Serial.flush();
    esp_deep_sleep_start(); // Everything below this line will **NEVER** be executed
}

void noise_loop() {
   for (int i=0; i < 60 && !INTE; i++) {
      itch_simulation();

      if (INTE) {
        break;
      }

      // small background vibration between simulation itches
      if (BACKGROUND_VIBRATION == 1) {
        ledcWrite(VIBRATION_PWM_CH, 75);
      }
      
      // Wait before next itch simulation
      delay(TIME_INTERVAL);
   }
    // end of background vibration
    if (BACKGROUND_VIBRATION == 1) {
      ledcWrite(VIBRATION_PWM_CH, 0);
    }
}

void itch_simulation() {
    // Number of leds, don't change this!
    const int N_LED = 4;
    
    for (int i=0; i < ITCH_SIM_TRACK_LEN && !INTE; i++) {
        // Enable all Sensors
        if (VIBRATION == 1) {
          ledcWrite(VIBRATION_PWM_CH, VIBRATIONS[i]);
        }
        if (BUZZER == 1) {
          ledcWrite(BUZZER_PWM_CH, BUZZERS[i]);
        }
        // Leds sensor are intermitted, with pause between them
        // this behaviour provides itself a timeinterval in which buzzer and vibration is on
        if (LED == 1) {
           for (int led = 0; led < N_LED; led++) {
              for (int i = 0; i < N_LED; i++) {
                  if (i == led) {
                    if (i == 0) 
                      leds.setLED(1, 65535, 0, 0);
                    if (i == 1) 
                      leds.setLED(0, 0, 65535, 0);
                    if (i == 2) 
                      leds.setLED(2, 0, 0, 65535);
                    if (i == 3)
                      leds.setLED(i, 65535, 65535, 65535);
                    
                  }
                      
                  else {
                    if (i == 0) 
                      leds.setLED(1, 0, 0, 0);
                    if (i == 1) 
                      leds.setLED(0, 0, 0, 0);
                    if (i == 2) 
                      leds.setLED(2, 0, 0, 0);
                    if (i == 3)
                      leds.setLED(i, 0, 0, 0);
                  }
              }
              leds.write();
              // Split timeinterval duration in N_LED subinterval, for each subinterval one LED is turned
              delay(ITCH_SIM_DURATION / (ITCH_SIM_TRACK_LEN * N_LED));
          }
        }
        // If leds are turned OFF you have to provide a timeinterval in which buzzer and vibration is on
        else {
          delay(ITCH_SIM_DURATION / ITCH_SIM_TRACK_LEN);
        }
  
        // Disable all sensors
        if (VIBRATION == 1) {
          ledcWrite(VIBRATION_PWM_CH, 0);
        }
        if (BUZZER == 1) {
          ledcWrite(BUZZER_PWM_CH, 0); 
        }
    }
}

int adc_to_mv(int r1, int r2, uint32_t value) {

  // 10 bit ADC, 11db attenuation
  // 0.140 V --> 0
  // 2.450 V --> 710

  const int cal_val = 710;
  const int cal_voltage = 2450;
  const int zero_voltage = 140;

  const int adc_voltage = value * (cal_voltage - zero_voltage) / 710 + zero_voltage;

  return adc_voltage * (r1 + r2) / r2;
}
